/* ICV196.h */

#ifndef _ICV196_h
#define _ICV196_h


#include <linux/sched.h>
#include "Z8536.h"

/*
 * Uncomment this if you want to use optional VMEchip2 features. This option
 * is not activated by default because only MVME16x boards have this chip. You
 * must also use Makefile.chip2 to compile the driver.
 */

/*#define VMECHIP2*/

#ifdef VMECHIP2
#	include "VMEchip2lib.h"
#endif


/*----------------------------------------------------------------------
 * Definitions
 *---------------------------------------------------------------------*/

#define		u32	unsigned long
#define		u16	unsigned short
#define		u8	unsigned char


/*----------------------------------------------------------------------
 * Constants
 *---------------------------------------------------------------------*/

/*ICV196 boards properties*/
#define		MAX_BOARDS	4		/*maximum number of ICV196 boards*/
#define		MAX_PORTS	12		/*number of ports per ICV196 board*/
#define		ICV196_SIZE	256		/*bytes occupation of one ICV196 board*/

/*ICV196 default config*/
#define		IO_BOARD_ADR	0xE17C04	/*default ICV196 address, the board
						 *is from 0xE17C0400 to 0xE17C04FF*/
#define		VME_IRQ		3		/*default VMEbus irq level*/
#define		BASE_VECTOR	0x60		/*default interrupt base vector*/

/*ioctl() cmd constants*/
#define		RD		0x11
#define		WR		0x12
#define		SLEEP		0x13

/*ioctl() arg.op constants*/
#define		DIO_BIT		0x21
#define		DIO_BYTE	0x22
#define		DIO_WORD	0x23
#define		DIO_ALL		0x24
#define		IO_DIR		0x25
#define		RISE		0x26
#define		FALL		0x27
#define		ANY		0x28
#define		HIGH		0x29
#define		LOW		0x2A


/*----------------------------------------------------------------------
 * Initialization parameters
 *---------------------------------------------------------------------*/

#ifdef VMECHIP2
static u32 	*chip2_adr	= (u32*) VMECHIP2_ADR;	/*VMEchip2 address*/
static int	adr_mod		= ADR_MODIFIER;		/*VMEchip2 Address Modifier
							 *should be 0x39 or 0x3D
							 *for ICV196 boards*/
#endif

/*ICV196 board's addresses*/
static u32 	*io_adr[MAX_BOARDS]	= {	(u32*) IO_BOARD_ADR,
						NULL,
						NULL,
						NULL	 };

static int	base_vector[MAX_BOARDS]	= {BASE_VECTOR,0,0,0};	/*irq base
								 *vectors*/
static int	vme_irq		= VME_IRQ;		/*VME bus irq*/


/*----------------------------------------------------------------------
 * Conversion tables for the ADAS ICV196 group registers and irq vectors
 *---------------------------------------------------------------------*/

/*
 * The group's registers are not implemented in order so we use an array
 * to make the conversion between group number and real address on the 
 * ICV196 board
 */
static u8	groupconv[12] = {1, 0, 3, 2, 5, 4, 7, 6, 9, 8, 11, 10};

/*
 * When you declare a base vector on Z8536 chip for port A or B, the channels
 * of these ports are affected with base vector + 2*channel_num_on_port. For
 * example if port A (channels 0-7) has a base vector 0x60 then channel 0 will
 * have vector 0x60, channel 1 -> 0x62, ...,channel 7 -> 0x7E. In order to use
 * a continuous array of irq vectors we give (base vector + 1) to port B
 * (channels 8-15) so channel 8 -> 0x61, ..., channel 15 -> 0x7F.
 * When an it occures, the channel number is thus given by itconv[base_vector - 
 * irq_vectors]
 */
static u8	itconv[16] = {0, 8, 1, 9, 2, 10, 3, 11, 4, 12, 5, 13, 6, 14, 7, 15};


/*----------------------------------------------------------------------
 * Structure of the ioctl() arguments. See ioctl() calls documentation
 * for more details
 *---------------------------------------------------------------------*/

struct icv196arg
{
	u8	op;	/*
			 *RD/WR -> DIO_BIT, DIO_BYTE, DIO_WORD, DIO_ALL or IO_DIR
			 *SLEEP -> RISE, FALL, ANY, HIGH or LOW 
			 */
	u8	num;	/*
			 *RD/WR | DIO_BIT -> 0 to 95
			 *RD/WR | DIO_BYTE -> 0 to 11
			 *RD/WR | DIO_WORD -> 0 to 10
			 *RD/WR | DIO_ALL -> don't care
			 *RD/WR | IO_DIR -> don't care
			 *SLEEP -> 0 to 15
			 */
	u8	val[MAX_PORTS];	/*
				 *RD/WR | DIO_BIT -> val[0]=0/1
				 *RD/WR | DIO_BYTE -> val[0]=0x00 to 0xFF
				 *RD/WR | DIO_WORD -> val[0]=0x00 to 0xFF
						      val[1]=0x00 to 0xFF
				 *RD/WR | DIO_ALL -> val[0-11]=0x00 to 0xFF
				 *RD/WR | IO_DIR -> val[0]=0x00 to 0xFF
						    val[1]=0x00 to 0x0F
				 *SLEEP -> don't care
				 */
	u32	it_timeout;	/*
				 *RD/WR -> don't care
				 *SLEEP -> 0 to 0xFFFFFFFF
				 */
};

typedef struct icv196arg ICV196_arg;


/*----------------------------------------------------------------------
 * Structures that map the ADAS ICV196 board interrupts controler
 * (Z8536_regs) and I/O registers (ICV196_regs)
 *---------------------------------------------------------------------*/

struct z8536regs
{
	u8	PORTC;		/*Zilog Z8536 port C register*/
	u8	pad1;		/*Non-addressable*/
	u8	PORTB;		/*Zilog Z8536 port B register*/
	u8	pad2;		/*Non-addressable*/
	u8	PORTA;		/*Zilog Z8536 port A register*/
	u8	pad3;		/*Non-addressable*/
	u8	CTL_REG;	/*Zilog Z8536 control register*/
	u8	pad4[57];	/*Non-addressable*/
	u8	CS_REG;		/*VME interrupt level WRITE_Only register*/
};

typedef struct z8536regs Z8536_REGS;


struct icv196regs
{
	u16		CLEAR;		/*Read or Write this register clear
					 *the board and reinitialize all ports
					 *on input*/
	u8		GR[12];		/*Input/Output registers*/	
	u16		DIR;		/*Ports direction Write_Only register*/
	u32		pad[28];	/*Non-addressable*/
	u8		nu;		/*Non-used*/
	Z8536_REGS	CIO;		/*Z8536 CIO chip registers*/
};

typedef struct icv196regs ICV196_REGS;


/*----------------------------------------------------------------------
 * Structure that keep the ADAS ICV196 board state (one per board)
 *---------------------------------------------------------------------*/

struct icv196state
{
	u16	DIRcopy;		/*ICV196 Write_Only DIR register copy*/
	u8	CScopy;			/*ICV196 Write_Only CS_REG (VME interrupt
					 *level) register copy*/
	u8	ITuse[16];		/*Present use of the 16 interrupt channels 
					 *(>0->in use, 0->not in use)*/
	u8	ITevent[16];		/*Present waiting events of the 16 interrupt
					 *channels (RISE, FALL, ANY, HIGH or LOW)*/
	u8	Minor;			/*Board's minor number*/
	u8	Base_Vector;		/*Board's base_vector*/
	struct semaphore Mutex;		/*Protect shared ressources of the board
					 *(in particular the previous ones)*/
	struct wait_queue *w_queue[16];	/*Interrupted tasks queues
				 	 *(one per interrupt channel)*/
};

typedef struct icv196state ICV196_STATE;



/*----------------------------------------------------------------------
 * Local variables
 *---------------------------------------------------------------------*/

static ICV196_REGS	*ICV196_regs[MAX_BOARDS];


/*----------------------------------------------------------------------
 * Z8536 CIO chip's access macros
 *---------------------------------------------------------------------*/

#define STATE0_CIO(board, garbage)			\
	garbage = (ICV196_regs[board]->CIO).CTL_REG;

#define WRITE_CIO(board, register, value)		\
{							\
	(ICV196_regs[board]->CIO).CTL_REG = register;	\
	(ICV196_regs[board]->CIO).CTL_REG = value;	\
}

#define READ_CIO(board, register, value)		\
{							\
	(ICV196_regs[board]->CIO).CTL_REG = register;	\
	value = (ICV196_regs[board]->CIO).CTL_REG;	\
}


/*----------------------------------------------------------------------
 * Functions prototypes 
 *---------------------------------------------------------------------*/
/*
static int Install_Interrupt(u8, u8);
static void Uninstall_Interrupt(u8);
static void ICV196_handler(int, void*, struct pt_regs*);
static int ICV196_init();
static int ICV196_ioctl (struct inode*, struct file*, int, void*);
static int ICV196_open (struct inode*, struct file*);
static void ICV196_release (struct inode*, struct file*);
int init_module ();
void cleanup_module (); 
*/


#endif	/* _ICV196_h */
