/* 
 * ICV196.c 
 *
 * ICV196 - kernel module DEVICE DRIVER for the ADAS ICV196 I/O board
 * by Integrated Real Time Systems (IRTS) - julien23@multimania.com
 * for European Synchrotron Radiation Facility (ESRF)
 *
 * OS : VME Linux/m68k 2.0 and 2.2
 *
 * VERSION : 1.1
 *
 * This device driver makes an easy access to the 96 digital I/O of the
 * ADAS ICV196 board via IOCTL() calls. You can also make your processes/
 * threads waiting for one of the 16 interruptible channels of the board
 * with ioctl() calls.
 *
 * 08/08/1999 (0.1) : Multi-boards support. 
 *
 * 10/08/1999 (0.2) : Enabled reading on output channels.
 *
 * 12/08/1999 (0.3) : Made the VMEchip2 features optional.
 *
 * 19/08/1999 (0.3.1) : Fixed different bugs.
 *
 * 20/08/1999 (0.4) : Added probe call for boards' addresses.
 *
 * 24/08/1999 (0.5) : Replaced edges waiting by levels waiting in order to
 *		      handle multiple interrupts at the same time on same port.
 *
 * 25/08/1999 (0.5.1) : Allowed both edges and levels waiting. User will have
 * 			be carefull with several SLEEP actions on edges.
 *			Replaced ITedge by ITevent and deleted it_edge from
 *			ICV196_arg (use op field now).
 *
 * 26/08/1999 (0.5.2) : Fixed bug with ^C wake_up.
 *
 * 25/11/1999 (1.1) : Fixed some warning.
 */



/*----------------------------------------------------------------------
 * Define constants
 *---------------------------------------------------------------------*/

#ifndef __KERNEL__		/*kernel optimization*/
#	define __KERNEL__
#endif

#ifndef MODULE			/*always a module*/
#	define MODULE
#endif



/*----------------------------------------------------------------------
 * Include files 
 *---------------------------------------------------------------------*/

#include <linux/config.h>	/*required for device drivers*/
#include <linux/module.h>	/*module optimization*/
#include <linux/version.h>	/*module needs to be in the good version*/ 
#include <linux/kernel.h>	/*kernel calls*/	
#include <linux/errno.h>	/*error flags*/
#include <linux/malloc.h>	/*pointers management*/
#include <linux/fs.h>		/*struct inode and struct file declaration*/
#include <linux/ioport.h>	/*check, request and release I/O region*/
#include <linux/wrapper.h>	/*sleeping functions*/

#include "ICV196.h"
#include "probelib.h"



/*----------------------------------------------------------------------
 * Macros 
 *---------------------------------------------------------------------*/

/*
 * These macros are used to determine which kernel you are using and to
 * compil the driver with the appropriate kernel calls
 */
#ifndef VERSION_CODE
#  define VERSION_CODE(vers,rel,seq) ( ((vers)<<16) | ((rel)<<8) | (seq) )
#endif

#if LINUX_VERSION_CODE < VERSION_CODE(2,0,0)	/*kernel < 2.0 - doesn't work*/
#  error "This kernel is too old: not supported by this file"
#endif
#if LINUX_VERSION_CODE < VERSION_CODE(2,1,0)
#  define LINUX_2_0	/*kernel 2.0*/
#else			/*kernel 2.1*/
#  include <asm/uaccess.h>	/*copy_to/from_user routine for kernel>2.1*/
#endif


/*
 * Uncomment the debug level you want to display on your console. You can also
 * see the debug messages by using 'dmesg' command on your shell.
 */
/*#define DEBUG*/	/*Uncomment DEBUG -> debug level 1 only*/
/*#define DEBUG2*/	/*Uncomment DEBUG2 -> debug level 2 only*/
			/*Uncomment both -> debug level 1 & 2*/

#ifdef DEBUG
#	define debug(x)		printk x
#else
#	define debug(x)
#endif

#ifdef DEBUG2
#	define debug2(x)	printk x
#else
#	define debug2(x)
#endif



/*----------------------------------------------------------------------
 * Local variables
 *---------------------------------------------------------------------*/

static u8		ICV196_boards	= 0;	/*number of boards (default->0)
						 *this number is fixed during
						 *the extract of parameters*/
static u16		ICV196_major	= 0;	/*0 (default) -> dynamique major
						 *allocation*/

static ICV196_STATE	*ICV196_state;	/*this struct is used to save important
					 *parameters about each board*/



/*----------------------------------------------------------------------
 * Function: Install_Interrupt()
 *
 * Check if the interrupt channel is already used and makes the config if
 * not. In case the channel is used, we can wait for it too if the configured
 * waiting event is good.
 * 'minor' indicates which board is concerned.
 *
 * OUTPUT :	0 -> install has succeed
 *		>0 -> install has failed (see asm/errno.h for value's means)
 *---------------------------------------------------------------------*/

static int Install_Interrupt(minor, channel, it_event)
u8 minor;
u8 channel;
u8 it_event;
{
	ICV196_STATE	*state = ICV196_state + minor;

/*
 * Check if the channel's group is configured as input
 */
	if ( (state->DIRcopy >> (channel/8)) & 1 ) 
		return -EINVAL;

/*
 * Check if channel is already used by a process which is sleeping and
 * configure it if not
 */
	if ( state->ITuse[channel] )	/*channel in use*/
	{
	/*
	 * Check if the present waiting event is good
	 */
		if ( it_event == state->ITevent[channel] )
			return 0;	/*yes, nothing more to do*/

		else			/*no, can't use this channel now*/
		{		
			debug(("\t\t+Interrupt channel already in use with another waiting event\n"));
printk("\t\t+Interrupt channel already in use with another waiting event\n");
			return -EINVAL;
		}
	}
	else	/*channel not in use so we have to configure it*/
	{
		u8	ppr, ptr, pmr, csreg, val;
		volatile u8	dummy;

		if ( channel < 8 )	/*which port is concerned?*/
		{
			ppr = PP_REG_A;
			ptr = PT_REG_A;
			pmr = PM_REG_A;
			csreg = CS_REG_A;
		}
		else
		{
			ppr = PP_REG_B;
			ptr = PT_REG_B;
			pmr = PM_REG_B;
			csreg = CS_REG_B;
		}

	/*
	 * Stop interrupts while programming the new channel
	 */
		STATE0_CIO(minor, dummy);	/*put the chip in state 0*/

		WRITE_CIO(minor, csreg, 0xE0);	/*clear it on port A/B while 
						 *programming*/
		switch (it_event)
		{
		case RISE:	/*bits value : ppr=1, ptr=1, pmr=1*/
			READ_CIO(minor, ppr, val);
			val |= (u8) (1 << (channel%8));
			WRITE_CIO(minor, ppr, val);

			READ_CIO(minor, ptr, val);
			val |= (u8) (1 << (channel%8));
			WRITE_CIO(minor, ptr, val);

			READ_CIO(minor, pmr, val);
			val |= (u8) (1 << (channel%8));
			WRITE_CIO(minor, pmr, val);

			break;
		
		case FALL:	/*bits value : ppr=0, ptr=1, pmr=1*/
			READ_CIO(minor, ppr, val);
			val &= (u8) ~(1 << (channel%8));
			WRITE_CIO(minor, ppr, val);

			READ_CIO(minor, ptr, val);
			val |= (u8) (1 << (channel%8));
			WRITE_CIO(minor, ptr, val);

			READ_CIO(minor, pmr, val);
			val |= (u8) (1 << (channel%8));
			WRITE_CIO(minor, pmr, val);

			break;

		case HIGH:	/*bits value : ppr=1, ptr=0, pmr=1*/
			READ_CIO(minor, ppr, val);
			val |= (u8) (1 << (channel%8));
			WRITE_CIO(minor, ppr, val);

			READ_CIO(minor, ptr, val);
			val &= (u8) ~(1 << (channel%8));
			WRITE_CIO(minor, ptr, val);

			READ_CIO(minor, pmr, val);
			val |= (u8) (1 << (channel%8));
			WRITE_CIO(minor, pmr, val);

			break;
		
		case LOW:	/*bits value : ppr=0, ptr=0, pmr=1*/
			READ_CIO(minor, ppr, val);
			val &= (u8) ~(1 << (channel%8));
			WRITE_CIO(minor, ppr, val);

			READ_CIO(minor, ptr, val);
			val &= (u8) ~(1 << (channel%8));
			WRITE_CIO(minor, ptr, val);

			READ_CIO(minor, pmr, val);
			val |= (u8) (1 << (channel%8));
			WRITE_CIO(minor, pmr, val);

			break;

		case ANY:	/*bits value : ppr=x, ptr=1, pmr=0*/
			READ_CIO(minor, ptr, val);
			val |= (u8) (1 << (channel%8));
			WRITE_CIO(minor, ptr, val);

			READ_CIO(minor, pmr, val);
			val &= (u8) ~(1 << (channel%8));
			WRITE_CIO(minor, pmr, val);

			break;
		}

	/*
	 * Save the present event waiting of the channel
	 */
		state->ITevent[channel] = it_event;

	/*
	 * Enable back the interrupts on ICV196
	 */ 
		WRITE_CIO(minor, csreg, 0x20);	/*clear interrupt pending on
						 *port A/B*/
		WRITE_CIO(minor, csreg, 0xC0);	/*enable interrupt port A/B*/
		STATE0_CIO(minor, dummy);	/*put the chip in state 0*/
		STATE0_CIO(minor, dummy);	/*put the chip in state 0*/

		return 0;
	}
}




/*----------------------------------------------------------------------
 * Function: Uninstall_Interrupt()
 *
 * This routine makes the interrupt channel inactive. We use it when the
 * interrupt has occured to disable this one until we need it again.
 * 'minor' indicates which board is concerned. 
 *---------------------------------------------------------------------*/

static void Uninstall_Interrupt(minor, channel)
u8 minor;
u8 channel;
{
	u8		val, ptr, pmr, csreg;
	volatile u8	dummy;

	if ( channel < 8 )	/*which port is concerned?*/
	{
		csreg = CS_REG_A;
		ptr = PT_REG_A;
		pmr = PM_REG_A;
	}
	else
	{
		csreg = CS_REG_B;
		ptr = PT_REG_B;
		pmr = PM_REG_B;
	}

	STATE0_CIO(minor, dummy);	/*put the chip in state 0*/

	READ_CIO(minor, ptr, val);
	val &= (u8) ~(1 << (channel%8));
	WRITE_CIO(minor, ptr, val);

	READ_CIO(minor, pmr, val);
	val &= (u8) ~(1 << (channel%8));
	WRITE_CIO(minor, pmr, val);

	WRITE_CIO(minor, csreg, 0x20);	/*clear interrupt pending on port A/B*/
	WRITE_CIO(minor, csreg, 0xC0);	/*enable interrupt port A/B*/
}



/*----------------------------------------------------------------------
 * Function: ICV196_handler()
 *
 * ICV196 Device Driver's interrupts handler routine. This routine wakes
 * up all the processes/threads which were sleeping, waiting for this
 * interrupt. Then it makes the channel uninstalled because it's no more
 * in use
 *---------------------------------------------------------------------*/

static void ICV196_handler(irq, dev_id, regs)
int irq;
void *dev_id;
struct pt_regs *regs;
{
	ICV196_STATE	*state = (ICV196_STATE *) dev_id;
	u8		channel = itconv[irq - (int)(state->Base_Vector)];

	debug(("  -ICV196_handler() :\n\t.IRQ vector = 0x%02X\n\t.channel = %d\n", irq, channel));

/*
 * Wake up all the processes waiting for this interrupt
 */
	wake_up(&(state->w_queue[channel]));

/*
 * Update the ITuse array
 */
	state->ITuse[channel] = 0;	/*0->no more used*/

/*
 * Disable the interrupt channel on the Z8536 chip and enable the interrupts
 * on port A/B because it has been disabled by the interrupt
 */
	Uninstall_Interrupt(state->Minor, channel);
}



/*----------------------------------------------------------------------
 * Function: ICV196_init()
 *
 * Makes the ICV196 Device Driver's initialization. It is called by the
 * init_module routine in order to check that hardware is well configured
 * and to reserve memory and IRQ vectors for the driver.
 *
 * OUTPUT :	0 -> init has succeed
 *		>0 -> init has failed (see asm/errno.h for value's means)
 *---------------------------------------------------------------------*/

static int ICV196_init()
{
	int		err, i, j, k, d;
	volatile u8	dummy;

#ifdef VMECHIP2
/*
 * Check if the VMEchip is well configured to deal with the ICV196 boards
 */
	for (i=0; i<ICV196_boards; i++)
	{
		err = VMEchip2_check_params(chip2_adr, adr_mod, io_adr[i], ICV196_SIZE);
		if ( err )	/*check has failed*/
		{
			printk("  -Module parameters :\n\t.chip2_adr = 0x%p\n\t.io_adr = 0x%p, 0x%p, 0x%p, 0x%p\n\t.adr_mod = 0x%02X\n\t.base_vector = 0x%02X, 0x%02X, 0x%02X, 0x%02X\n\t.vme_irq = %d\n", chip2_adr, io_adr[0], io_adr[1], io_adr[2], io_adr[3], adr_mod, base_vector[0], base_vector[1], base_vector[2], base_vector[3], vme_irq); 
			VMEchip2_status(chip2_adr);	/* Dump the VMEchip2's LCSRs*/
			goto std_error;
		}
	}
#endif

/*
 * Check if the I/O regions you request are not already used by a device
 */
	for (i=0; i<ICV196_boards; i++)
	{
		err = check_region((u32)io_adr[i], ICV196_SIZE);
		if ( err )
		{
			printk("  -The IO address space you indicate is already used\n");
			goto region_error; /*may be some region has already been
					*requested and 'i' is the number of those*/
		}
		request_region((u32)io_adr[i], ICV196_SIZE, "ICV196");
	}

/*
 * Allocate the memory for ICV196_state stucture
 */
	ICV196_state = (ICV196_STATE *) kmalloc(sizeof(ICV196_STATE) * ICV196_boards, GFP_KERNEL);
	if ( ICV196_state == NULL )
	{
		printk("  -Unable to get enough memory to drive the ICV196 device");
		err = -ENOMEM;
		i = ICV196_boards;	/*all the regions have been requested*/
		goto region_error;	/*region has already been requested!*/
	}

	/*set all the ICV196_state structure to 0*/
	memset(ICV196_state, 0, sizeof(ICV196_STATE) * ICV196_boards);

/*
 * Reset the board and set the init values of structs
 */
	for (i=0; i<ICV196_boards; i++)
	{
		(ICV196_state+i)->Minor = i;	/*save each board's minor
						 *number (0, 1, 2, 3)*/
		(ICV196_state+i)->Base_Vector = base_vector[i];	/*save the base
								 *vector*/
		(ICV196_state+i)->Mutex = MUTEX; /*protect the access and the
						  *back-up values of the board*/
		ICV196_regs[i] = (ICV196_REGS *) io_adr[i];  /*map the board*/
		ICV196_regs[i]->CLEAR = 0;	/*read or write this register
						 *reset the board*/
	}

#ifdef VMECHIP2
/*
 * Configure the VMEchip2's VME irq level given by 'vme_irq'
 */
	err = VMEchip2_irq(chip2_adr, CONFIG, vme_irq);
	if ( err )	/*irq init on VMEchip2 has failed*/
	{
		printk("  -Unable to initialize irq on VMEchip2\n");
		goto malloc_error;	/*malloc has been done*/
	}

/*
 * Disable the VMEchip2's VME irq level given by 'vme_irq' until we need
 * it (end of initialization)
 */
	err = VMEchip2_irq(chip2_adr, DISABLE, vme_irq);
	if ( err )	/*disabling irq on VMEchip2 has failed*/
	{
		printk("  -Unable to disable irq on VMEchip2\n");
		goto malloc_error;	/*malloc has been done*/
	}
#endif

/*
 * Configure the ICV196's VME irq level
 */
	for (i=0; i<ICV196_boards; i++)
	{
		ICV196_regs[i]->CIO.CS_REG = ~(u8)(1 << vme_irq);
		(ICV196_state+i)->CScopy = ~(u8)(1 << vme_irq);
	}

/*
 * Install the interrupt handler on all 16 vectors of each board
 */
	for (i=0; i<ICV196_boards; i++)
	{
		ICV196_STATE	*state = (ICV196_state + i);
		for (d=0; d<16; d++)
		{
			err = request_irq(state->Base_Vector+d, ICV196_handler, 0, "ICV196", state);
			if ( err )	/*one request has failed*/
			{
				printk("  -Interrupt \"ICV196(%d)\" : cannot claim irq vector 0x%02X\n", itconv[d], state->Base_Vector+d);
				goto irq_error;	/*may be some irqs have been affected*/
			}
		}
	}

/*
 * Configure the Z8536 CIO chips in order to deal with interrupts
 */
	for (i=0; i<ICV196_boards; i++)
	{
	ICV196_STATE	*state = (ICV196_state + i);

	STATE0_CIO(i, dummy);			/*put the chip in state 0*/
	WRITE_CIO(i, MIC_REG, 0x01);		/*set the reset bit*/ 
	(ICV196_regs[i]->CIO).CTL_REG = 0x00;	/*clear the reset bit (0) before
					 *programming the other control bits*/
	WRITE_CIO(i, IV_REG_A, state->Base_Vector);   /*port A -> even irq vectors*/
	WRITE_CIO(i, IV_REG_B, state->Base_Vector+1); /*port B -> odd irq vectors*/

	WRITE_CIO(i, MCC_REG, 0x84);	/*enable ports A and B*/

	WRITE_CIO(i, SC_REG_A, 0x00);	/*normal input/output on port A*/
	WRITE_CIO(i, CS_REG_A, 0xE0);	/*clear it on port A*/

	WRITE_CIO(i, SC_REG_B, 0x00);	/*normal input/output on port B*/
	WRITE_CIO(i, CS_REG_B, 0xE0);	/*clear it on port B*/

	WRITE_CIO(i, MIC_REG, 0x98);	/*ports A and B it vectors include*/

	WRITE_CIO(i, MS_REG_A, 0x1F);	/*port A/B bit port and OR 
					 *priority encoded vector mode*/
	WRITE_CIO(i, DPP_REG_A, 0x00);	/*port A/B non inverting input*/
	WRITE_CIO(i, DD_REG_A, 0xFF);	/*init port A/B as input*/

	WRITE_CIO(i, MS_REG_B, 0x1F);	/*port B bit port and OR 
					 *priority encoded vector mode*/
	WRITE_CIO(i, DPP_REG_B, 0x00);	/*port B non inverting input*/
	WRITE_CIO(i, DD_REG_B, 0xFF);	/*init port B as input*/

	STATE0_CIO(i, dummy);		/*put the chip in state 0*/
	}

#ifdef VMECHIP2
/*
 * Now we can enable the VMEchip2's VME irq level given by 'vme_irq'
 */
	err = VMEchip2_irq(chip2_adr, ENABLE, vme_irq);
	if ( err ) 	/*Enabling irq on VMEchip2 has failed*/
	{
		printk("  -Unable to enable irq on VMEchip2\n");
		i = ICV196_boards;	/*set the cleanning parameters..*/
		d = 16;			/*..for free_irq*/
		goto irq_error;		/*irqs have been affected*/
	}
#endif

/*
 * If everything have succedded we can continue
 */
	return 0;

/*
 * Something has failed
 */
	irq_error:	/*'i' contains the board number where request has failed
			 *and 'd' the offset number so we have to free all the
			 *already requested ones*/
		for (k=i; k>=0; k--)
		{
			ICV196_STATE *state = (ICV196_state + k);

			for (j=0; j<d; j++)
				free_irq(state->Base_Vector+j, state);
			d=16;
		}

#ifdef VMECHIP2		/*this label is only used with VMEchip2 features*/
	malloc_error:
#endif
		kfree(ICV196_state);
		i = ICV196_boards; /*here all the region have already been requested*/

	region_error:	/*'i' contains the board number where request has failed
			 *the former ones have already been requested->release*/
		for (k=0; k<i; k++)
			release_region((u32)io_adr[k], ICV196_SIZE);

#ifdef VMECHIP2		/*this label is only used with VMEchip2 features*/
	std_error:
#endif
		return err;
}



/*----------------------------------------------------------------------
 * Function: ICV196_ioctl()
 *
 * Handle the ioctl() calls. This routine realizes the most part of the
 * I/O Device Driver's work. It is the only way to make an access to the
 * ICV196 I/O board.
 *
 * INPUT:	FD -> file descriptor
 *		CMD -> RD, WR or SLEEP
 *		ARG  .OP -> (RD/WR) DIO_BIT, DIO_BYTE, DIO_WORD, DIO_ALL or IO_DIR
 *			    (SLEEP) RISE, FALL, ANY, HIGH or LOW
 *		     .NUM -> bit (0-95), group (0-11), word (0-10) or
 *		      interrupt channel (0-16) number
 *		     .VAL -> bit, byte, word, all value
 *		     .IT_TIMEOUT -> waiting timeout (0 for none)
 *
 * OUTPUT :	0 -> command has succeed
 *		>0 -> command has failed (see asm/errno.h for value's means)
 *---------------------------------------------------------------------*/

static int ICV196_ioctl (inode, file, cmd, arg)
struct inode *inode;
struct file *file;
int cmd;
void *arg; 
{
	u8		minor;
	ICV196_arg 	_arg;
	ICV196_STATE	*state;
	ICV196_REGS	*regs;

/*
 * Determine which board is going to be accessed and then initialize
 * the local pointers that give us access to the board status and registers.
 * The minor number indicates which board is concerned : 0 for the first board
 * ... 3 for the last one (fourth)
 */
	minor = MINOR(inode->i_rdev);	/*get minor*/

	if ( minor >= ICV196_boards )	/*is it a correct minor?*/
	{
		debug(("  -ICV196_ioctl() : Bad minor number"));
		return -ENODEV;
	}

	state = ICV196_state + minor;	/*local pointer to status struct*/
	regs = ICV196_regs[minor];	/*local pointer to board's registers struct*/


	debug2(("  -ICV196_ioctl() : board %d\n\t.cmd = 0x%X\n\t.arg = 0x%p\n", minor, cmd, arg));

/*
 * Copy argument from user memory space to kernel memory space. We have to
 * do that to get user arguments. The routine has been optimized on kernel
 * 2.1 and higher
 */
#ifdef LINUX_2_0
	if (verify_area(VERIFY_READ, arg, sizeof(ICV196_arg)) == -EFAULT)
		return -EFAULT;
	memcpy_fromfs (&_arg, arg, sizeof(ICV196_arg));
#else
	if ( copy_from_user(&_arg, arg, sizeof(ICV196_arg)) )
		return -EFAULT;
#endif

/*
 * Dump user's parameters when DEBUG2
 */
#ifdef DEBUG2
	{
		int i;
		printk("\t\t+op = %#X\n\t\t+num = %d\n\t\t+val = 0x", _arg.op, _arg.num);
		for (i=0; i<strlen(_arg.val); i++)
			printk("%02X", _arg.val[0]);
		printk("\n\t\t+it_timeout = %ld\n", _arg.it_timeout);
	}
#endif

	switch (cmd)
	{
	case RD:
		switch (_arg.op)
		{
		case IO_DIR:
			debug(("\t.RD | DIRECTION\n"));

		/*
		 * We must protect all the shared ressources like that
		 */	
			down( &(state->Mutex) );
			_arg.val[0] = (u8) (state->DIRcopy & 0x00FF);
			_arg.val[1] = (u8) ((state->DIRcopy & 0xFF00)>>8);
			up( &(state->Mutex) );

/*
 * Copy back the arg structure that contain the result of the read operation
 */
#ifdef LINUX_2_0
			memcpy_tofs (arg, &_arg, sizeof(ICV196_arg));
#else
			if ( copy_to_user(arg, &_arg, sizeof(ICV196_arg)) )
				return -EFAULT;
#endif

			debug2(("\t\t+DIRcopy = 0x%04X\n\t\t+Done\n", state->DIRcopy));
			return 0;

		case DIO_BIT:
			debug(("\t.RD | DIO_BIT\n"));
			debug2(("\t\t+DIRcopy = %04X\n", state->DIRcopy));
			
			down( &(state->Mutex) );
		/*
		 * Check if channel is from 0 to 95 and if its group is input
		 */
			if (	(_arg.num <= 95)
				/*
				&& !((state->DIRcopy>>(_arg.num/8)) & 1)
				 *
				 * This comment enables user to read output channels
				 * It is however imperative to wait 15s between
				 * the output instruction and readback of the
				 * same channel so that the propagation time to
				 * switch the open collectors is reached
				 */
									) 
			{
				_arg.val[0] = ((regs->GR[groupconv[_arg.num/8]] >> (_arg.num%8)) & 1); 
				up( &(state->Mutex) );

#ifdef LINUX_2_0
				memcpy_tofs (arg, &_arg, sizeof(ICV196_arg));
#else
				if ( copy_to_user(arg, &_arg, sizeof(ICV196_arg)) )
					return -EFAULT;
#endif

				debug(("\t\t+Done\n"));
				return 0;
			}
			else
			{
				up( &(state->Mutex) );
				debug(("\t\t+Bad RD | DIO_BIT arguments - bit number must be from 0 to 95\n"));
				return -EINVAL;
			}	

		case DIO_BYTE:
			debug(("\t.RD | DIO_BYTE\n"));
			debug2(("\t\t+DIRcopy = %04X\n", state->DIRcopy));

			down( &(state->Mutex) );
		/*
		 * Check if group is from 0 to 11 and if its way is input
		 */
			if (	(_arg.num <= (MAX_PORTS-1)) 
				/*
				&& !((state->DIRcopy>>_arg.num) & 1)
				 *
				 * See RD | DIO_BIT for more informations
				 */
									)
			{
				_arg.val[0] = regs->GR[groupconv[_arg.num]]; 
				up( &(state->Mutex) );

#ifdef LINUX_2_0
				memcpy_tofs (arg, &_arg, sizeof(ICV196_arg));
#else
				if ( copy_to_user(arg, &_arg, sizeof(ICV196_arg)) )
					return -EFAULT;
#endif

				debug(("\t\t+Done\n"));
				return 0;
			}
			else
			{
				up( &(state->Mutex) );
				debug(("\t\t+Bad RD | DIO_BYTE arguments - byte number must be from 0 to 11\n"));
				return -EINVAL;
			}	

		case DIO_WORD:
			debug(("\t.RD | DIO_WORD\n"));
			debug2(("\t\t+DIRcopy = %04X\n", state->DIRcopy));
			
			down( &(state->Mutex) );
		/*
		 * Check if word is from 0 to 10 and if its 2 groups are input
		 */
			if (	(_arg.num <= 10)
				/*
				&& (((state->DIRcopy>>_arg.num) & 3) == 0)
				 *
				 * See RD | DIO_BIT for more informations
				 */
									)
			{
				_arg.val[0] = regs->GR[groupconv[_arg.num]]; 
				_arg.val[1] = regs->GR[groupconv[_arg.num+1]]; 
				up( &(state->Mutex) );

#ifdef LINUX_2_0
				memcpy_tofs (arg, &_arg, sizeof(ICV196_arg));
#else
				if ( copy_to_user(arg, &_arg, sizeof(ICV196_arg)) )
					return -EFAULT;
#endif

				debug(("\t\t+Done\n"));
				return 0;
			}
			else
			{
				up( &(state->Mutex) );
				debug(("\t\t+Bad RD | DIO_WORD arguments - check the groups config (READ=0)\n"));
				return -EINVAL;
			}	

		case DIO_ALL:
			debug(("\t.RD | DIO_ALL\n"));
			debug2(("\t\t+DIRcopy = %04X\n", state->DIRcopy));
			
			down( &(state->Mutex) );
		/*
		 * Check if all the groups are input
		 */
			/*
			if ( (state->DIRcopy & 0xFFF) == 0 ) 
			 *
			 * No more test if we enable input and output reading
			 * See RD | DIO_BIT for more informations
			 */
			{
				int i;

				for (i=0; i<12; i++)
					_arg.val[i] = regs->GR[groupconv[i]]; 
				up( &(state->Mutex) );

#ifdef LINUX_2_0
				memcpy_tofs (arg, &_arg, sizeof(ICV196_arg));
#else
				if ( copy_to_user(arg, &_arg, sizeof(ICV196_arg)) )
					return -EFAULT;
#endif

				debug(("\t\t+Done\n"));
				return 0;
			}
			/*
			else
			{
				up( &(state->Mutex) );
				debug(("\t\t+Bad RD | DIO_ALL arguments - check the groups config (READ=0)\n"));
				return -EINVAL;
			}	
			*/

		default:
			debug(("\t\t+Bad RD arg.op\n"));
			return -EINVAL;
		}

	case WR:
		switch (_arg.op)
		{
			u16 dir;

		case IO_DIR:
			debug(("\t.WR | DIRECTION\n"));

			dir = (u16)_arg.val[0] + (((u16)_arg.val[1])<<8);
			down( &(state->Mutex) );
			regs->DIR = dir; 
			state->DIRcopy = dir;
			up( &(state->Mutex) );

			debug2(("\t\t+DIRcopy = 0x%04X\n", dir));
			debug(("\t\t+Done\n"));
			return 0;

		case DIO_BIT:
			debug(("\t.WR | DIO_BIT\n"));
			debug2(("\t\t+DIRcopy = %04X\n", state->DIRcopy));
			
			down( &(state->Mutex) );
		/*
		 * Check if channel is from 0 to 95, if its group is output and
		 * if value is 0 or 1
		 */
			if (	(_arg.num<=95) && 
				((state->DIRcopy>>(_arg.num/8)) & 1) &&
				(_arg.val[0]==0 || _arg.val[0]==1) )
			{
				u8 dummy = regs->GR[groupconv[_arg.num/8]];

				if (_arg.val[0]==0)
					dummy &= ~(1 << (_arg.num%8));
				else
					dummy |= (1 << (_arg.num%8));

				regs->GR[groupconv[_arg.num/8]] = dummy;
				up( &(state->Mutex) );
					
				debug(("\t\t+Done\n"));
				return 0;
			}
			else
			{
				up( &(state->Mutex) );
				debug(("\t\t+Bad WR | DIO_BIT arguments - check the group config (WRITE=1)\n"));
				return -EINVAL;
			}

		case DIO_BYTE:
			debug(("\t.WR | DIO_BYTE\n"));
			debug2(("\t\t+DIRcopy = %04X\n", state->DIRcopy));
			
			down( &(state->Mutex) );
		/*
		 * Check if group is from 0 to 11 and is output
		 */
			if (	(_arg.num<=(MAX_PORTS-1)) && 
				((state->DIRcopy>>_arg.num) & 1) ) 
			{
				regs->GR[groupconv[_arg.num]] = _arg.val[0];
				up( &(state->Mutex) );
					
				debug(("\t\t+Done\n"));
				return 0;
			}
			else
			{
				up( &(state->Mutex) );
				debug(("\t\t+Bad WR | DIO_BYTE arguments - check the group config (WRITE=1)\n"));
				return -EINVAL;
			}

		case DIO_WORD:
			debug(("\t.WR | DIO_WORD\n"));
			debug2(("\t\t+DIRcopy = %04X\n", state->DIRcopy));
			
			down( &(state->Mutex) );
		/*
		 * Check if word is from 0 to 10 and if its 2 groups are output
		 */
			if (	(_arg.num<=10) && 
				(((state->DIRcopy>>_arg.num) & 3) == 3) )
			{
				regs->GR[groupconv[_arg.num]] = _arg.val[0];
				regs->GR[groupconv[_arg.num+1]] = _arg.val[1];
				up( &(state->Mutex) );
					
				debug(("\t\t+Done\n"));
				return 0;
			}
			else
			{
				up( &(state->Mutex) );
				debug(("\t\t+Bad WR | DIO_WORD arguments - check the groups config (WRITE=1)\n"));
				return -EINVAL;
			}

		case DIO_ALL:
			debug(("\t.WR | DIO_ALL\n"));
			debug2(("\t\t+DIRcopy = %04X\n", state->DIRcopy));
			
			down( &(state->Mutex) );
		/*
		 * Check is all the 12 groups are output
		 */
			if ( (state->DIRcopy & 0xFFF) == 0xFFF ) 
			{
				int i;

				for (i=0; i<12; i++)
					regs->GR[groupconv[i]] = _arg.val[i];
				up( &(state->Mutex) );
					
				debug(("\t\t+Done\n"));
				return 0;
			}
			else
			{
				up( &(state->Mutex) );
				debug(("\t\t+Bad WR | DIO_ALL arguments - check the groups config (WRITE=1)\n"));
				return -EINVAL;
			}

		default:
			debug(("\t\t+Bad WR arg.op\n"));
			return -EINVAL;
		}

	case SLEEP:
		debug(("\t.SLEEP\n"));
		debug2(("\t\t+DIRcopy = %04X\n", state->DIRcopy));

		if ( _arg.num > 15 )
		{
			debug(("\t\t+Bad SLEEP arg.num\n"));
			return -EINVAL;
		}

		switch (_arg.op)
		{
			int err;

		case RISE:
		case FALL:
		case ANY:
		case HIGH:
		case LOW:

		/*
		 * This part is uninterruptible because we touch Z8536 config
		 * such as the interrupt handler. The interrupt are enabled back
		 * by the driver when the install has failed or by the sleep_on
		 * routine when the process/thread goes to sleep
		 */ 
			cli();
			err = Install_Interrupt(state->Minor, _arg.num, _arg.op);
			if ( err )
			{
				sti();
				debug(("\t\t+Install_Interrupt() has failed\n"));
				return err;
			}

		/*
		 * One more process is going to wait for the channel
		 */
			state->ITuse[_arg.num]++;

		/*
		 * If the timeout specified by user is not 0, we use an 
		 * interruptible_sleep_on with the correct timeout.
		 * If it is 0, we also use interruptible_sleep_on instead of
		 * sleep_on to avoid the process/thread to sleep all the time
		 * (even kill -9 doesn't stop it with sleep_on) in case of
		 * problem. In kernel 2.0 we have to put 0 as a timeout.
		 * Things are clearer on the 2.1 kernels or higher because there
		 * is now 4 functions : sleep_on, interruptible_sleep_on,
		 * sleep_on_timeout and interruptible_sleep_on_timeout
		 */
			if ( _arg.it_timeout )	/*sleep_on with timeout*/
			{
				u32 timeout;

				debug(("\t\t+Interruptible_sleep_on : timeout = %ld\n", _arg.it_timeout));

#ifdef LINUX_2_0
				current->timeout = jiffies + _arg.it_timeout;
				interruptible_sleep_on(&(state->w_queue[_arg.num]));
				cli();
				timeout = current->timeout;
#else
				timeout = interruptible_sleep_on_timeout(&(state->w_queue[_arg.num]), jiffies + _arg.it_timeout);
				cli();
#endif

				if ( timeout == 0 )	/*woken up by timeout*/
				{
					debug(("  -Woken_up (PID = %d) by timeout\n", current->pid));

					if ( (--state->ITuse[_arg.num]) == 0 )
						Uninstall_Interrupt(state->Minor ,_arg.num);
					sti();
				}
				else
				{
#ifdef LINUX_2_0
					if ( current->signal & ~current->blocked )
#else
					if ( signal_pending(current) )
#endif
					/*woken up by signal from user*/
					{
						debug(("  -Woken_up (PID = %d) by signal\n", current->pid));

						if (state->ITuse[_arg.num] != 0)
							state->ITuse[_arg.num]--;

						if (state->ITuse[_arg.num] == 0)
							Uninstall_Interrupt(state->Minor ,_arg.num);

						sti();

						return -ERESTARTSYS;
					}

					/*woken up by the interrupt*/
					sti();
					debug(("  -Woken_up (PID = %d) by interrupt\n", current->pid));
				}
			}
			else	/*sleep_on without timeout*/
			{
				debug(("\t\t+Sleep_on (no timeout)\n"));

#ifdef LINUX_2_0
				current->timeout = 0;
#endif

				interruptible_sleep_on(&(state->w_queue[_arg.num]));
				cli();

#ifdef LINUX_2_0
				if ( current->signal & ~current->blocked )
#else
				if ( signal_pending(current) )
#endif
				/*woken up by signal from user*/
				{
					debug(("  -Woken_up (PID = %d) by signal\n", current->pid));

					if (state->ITuse[_arg.num] != 0)
						state->ITuse[_arg.num]--;

					if (state->ITuse[_arg.num] == 0)
						Uninstall_Interrupt(state->Minor ,_arg.num);

					sti();

					return -ERESTARTSYS;
				}

				/*woken up by the interrupt*/
				sti();
				debug(("  -Woken_up (PID = %d) by interrupt\n", current->pid));
			}

			return 0;
	
		default:
			debug(("\t\t+Bad SLEEP arg.op\n"));
			return -EINVAL;
		}

	default:
		debug(("\t.Bad ioctl() command\n"));
		return -EINVAL;

	}
}



/*----------------------------------------------------------------------
 * Function: ICV196_open()
 *
 * Called when /dev/ICV196_x is opened. Update the usage count so the module
 * doesn't get unloaded while it is in use. 
 *---------------------------------------------------------------------*/

static int ICV196_open (inode, file)
struct inode *inode;
struct file *file;
{
	debug(("  -ICV196_open() : PID = %d\n", current->pid));

	MOD_INC_USE_COUNT;	/*one more process/thread plays with us*/
	return 0;
}



/*----------------------------------------------------------------------
 * Function: ICV196_release()
 *
 * Called when /dev/ICV196_x is closed or when the process exits. Update the
 * usage count so that the module get unloaded if no more process need it
 * in a certain amount of time. In kernel 2.1 or higher this function
 * return a value (alowed to fail now?)
 *---------------------------------------------------------------------*/
#ifdef LINUX_2_0
static void ICV196_release(inode, file)
#else
static int ICV196_release(inode, file)
#endif
struct inode *inode;
struct file *file;
{
	debug(("  -ICV196_release() : PID = %d\n", current->pid));

	MOD_DEC_USE_COUNT;	 /*one more process/thread leaves the game*/

#ifdef LINUX_2_0
	/*nothing to return*/
#else
	return 0;
#endif
}



/*----------------------------------------------------------------------
 * Device driver entries points 
 *---------------------------------------------------------------------*/

struct file_operations ICV196_fops =
{
	NULL,			/* _seek	*/
	NULL,			/* _read	*/
	NULL,			/* _write	*/
	NULL,			/* _readdir	*/
	NULL,			/* _select	*/
	ICV196_ioctl,		/* _ioctl	*/
	NULL,			/* _mmap	*/
	ICV196_open,		/* _open	*/
	ICV196_release,		/* _release	*/
};



/*----------------------------------------------------------------------
 * Function: init_module()
 *
 * Initialize the module
 *---------------------------------------------------------------------*/

int init_module()
{
	int err, err1, i=0;

	printk(KERN_INFO "ICV196 Device Driver\n");

/*
 * Extract and check loading parameters. If base_vector has not been specified,
 * we use a continuous irq vectors area (there is 16 vectors per board). Only
 * the 24 most significant bits are loaded with insmod so we have to make a
 * correct shift on the left.
 */
	while ( (i < MAX_BOARDS) && (io_adr[i] != NULL) )
	{
		io_adr[i] = (u32 *)((u32)io_adr[i] << 8);

	/*
	 * Probe the ICV196 board by reading at the first and the last addresses
	 * of the board
	 */
		err = probe((char*)io_adr[i], sizeof(short), -1);
		err1 = probe(((char*)io_adr[i])+ICV196_SIZE-2, sizeof(short), -1);
		if (err || err1)
		{
			printk("Board %d is not at the specified address\n", i);
			return -EFAULT;
		}

	/*
	 * Set default base vector if not specified
	 */
		if ( base_vector[i] == 0 )
			base_vector[i] = base_vector[i-1] + 16;

		ICV196_boards++;
		i++;
	}

	if ( (vme_irq <= 0) || (vme_irq > 7) )
	{
		printk("  -Bad VME IRQ level\n");
		return -EINVAL;
	}

/*
 * Register the ICV196 (char) Device Driver with the ICV196_major number. If
 * ICV196_major is set to 0 then the allocation number will be determined by
 * the kernel
 */
	err = register_chrdev(ICV196_major, "ICV196", &ICV196_fops);
	if ( err < 0 )	/*register_chrdev has failed*/
	{
		printk("  -Unable to get major for ICV196 device");
		return err;
	}
	if ( ICV196_major == 0)		/*dynamic major allocation*/
		ICV196_major = err;

/*
 * If error has appeared during ICV196_init, we quit with error
 */
	err = ICV196_init();
	if ( err )
	{
		unregister_chrdev(ICV196_major, "ICV196");
		return err;
	}

	debug(("  -%d boards have been successfully initialized\n", ICV196_boards));

	return 0;
}



/*----------------------------------------------------------------------
 * Function: cleanup_module()
 *
 * Undid whatever init_module did
 *---------------------------------------------------------------------*/

void cleanup_module()
{
	int i, j;

/*
 * Free all irq vectors
 */
	for (i=0; i<ICV196_boards; i++)
	{
		ICV196_STATE *state = (ICV196_state + i);

		for (j=0; j<16; j++)
			free_irq(state->Base_Vector+j, state);
	}

/*
 * Release the allocated memory
 */
	kfree(ICV196_state);

/*
 * Release the I/O board address region
 */
	for (i=0; i<ICV196_boards; i++)
		release_region((u32)io_adr[i], ICV196_SIZE);

 /*
  * Unregister the device before leaving
  */
	unregister_chrdev(ICV196_major, "ICV196");

	printk(KERN_INFO "End of ICV196 Device Driver\n\n");
}
