/* 
 * probe_c.c 
 *
 * PROBE - kernel module
 * by Integrated Real Time Systems (IRTS) - julien23@multimania.com
 * for European Synchrotron Radiation Facility (ESRF)
 *
 * OS : Linux/m68k 2.0 and 2.2 (has only been tested on 2.0)
 *
 * PROCESSOR : Motorola 680[2-4]0 (has only been tested on 68040)
 *
 * VERSION : 1.0
 *
 * This module provides a probe() function for kernel functions such as device
 * drivers or kernel modules. This is very usefull to avoid system crash when
 * you first try to access device's addresses
 *
 */




/*----------------------------------------------------------------------
 * Define constants
 *---------------------------------------------------------------------*/

#ifndef __KERNEL__		/*kernel optimization*/
#	define __KERNEL__
#endif

#ifndef MODULE			/*always a module*/
#	define MODULE
#endif



/*----------------------------------------------------------------------
 * Include files 
 *---------------------------------------------------------------------*/

#include <linux/config.h>	/*required for device drivers*/
#include <linux/module.h>	/*module optimization*/
#include <linux/version.h>	/*module needs to be in the good version*/ 
#include <linux/kernel.h>	/*kernel calls*/	
#include <linux/malloc.h>	/*pointers management*/
#include <linux/fs.h>		/*struct inode and struct file declaration*/

#include "probe.h"



/*----------------------------------------------------------------------
 * Macros 
 *---------------------------------------------------------------------*/

/*
 * These macros are used to determine which kernel you are using and to
 * compil the driver with the appropriate kernel calls
 */
#ifndef VERSION_CODE
#  define VERSION_CODE(vers,rel,seq) ( ((vers)<<16) | ((rel)<<8) | (seq) )
#endif

#if LINUX_VERSION_CODE < VERSION_CODE(2,0,0)	/*kernel < 2.0 - doesn't work*/
#  error "This kernel is too old: not supported by this file"
#endif
#if LINUX_VERSION_CODE < VERSION_CODE(2,1,0)
#  define LINUX_2_0	/*kernel 2.0*/
#else			/*kernel 2.1*/
#endif


/*
 * Uncomment DEBUG if you want to display comments on your console. You can 
 * also see the debug messages by using 'dmesg' command on your shell.
 */
#define DEBUG/**/	/*Uncomment DEBUG -> debug level 1 only*/

#ifdef DEBUG
#	define debug(x)		printk x
#else
#	define debug(x)
#endif



/*----------------------------------------------------------------------
 * Function: probe_open()
 *
 *---------------------------------------------------------------------*/

static int probe_open (inode, file)
struct inode *inode;
struct file *file;
{
	MOD_INC_USE_COUNT;	/*one more process/thread play with us*/
	return 0;
}



/*----------------------------------------------------------------------
 * Function: ICV196_release()
 *
 *---------------------------------------------------------------------*/
#ifdef LINUX_2_0
static void probe_release(inode, file)
#else
static int probe_release(inode, file)
#endif
struct inode *inode;
struct file *file;
{
	MOD_DEC_USE_COUNT;	 /*one more process/thread leave the game*/

#ifdef LINUX_2_0
	/*nothing to return*/
#else
	return 0;
#endif
}



/*----------------------------------------------------------------------
 * Device driver entries points 
 *---------------------------------------------------------------------*/

struct file_operations probe_fops =
{
	NULL,			/* _seek	*/
	NULL,			/* _read	*/
	NULL,			/* _write	*/
	NULL,			/* _readdir	*/
	NULL,			/* _select	*/
	NULL,			/* _ioctl	*/
	NULL,			/* _mmap	*/
	probe_open,		/* _open	*/
	probe_release,		/* _release	*/
};



/*----------------------------------------------------------------------
 * Function: cache_disable()
 *
 * Disable the cache control in order to access to real memory. Use the
 * asm_cache_disable routine to do it.
 *---------------------------------------------------------------------*/
static void cache_disable()
{
	unsigned long ps;

	save_flags(ps);
	cli();

	asm_cache_disable();

	restore_flags(ps);
}



/*----------------------------------------------------------------------
 * Function: cache_enable()
 *
 * Enable back the cache control. Use the asm_cache_enable routine to do
 * it.
 *---------------------------------------------------------------------*/
static void cache_enable()
{
	unsigned long ps;

	save_flags(ps);
	cli();

	asm_cache_enable();

	restore_flags(ps);
}



/*----------------------------------------------------------------------
 * Function: catchbuserr()
 *
 * Catch bus error. Ensure instruction is not re-run.
 *---------------------------------------------------------------------*/
void catchbuserr(efp)
register struct exframe *efp;
{
    caughtbuserr = 1;

#ifndef MC68040
	/* 
	 * In the 680[2-3]0, we reset the RC and RB flags of the special
	 * status word to prevent the bus cycle from being re-run.
	 * We'll also reset the DF flag just in case.
	 */

	efp->ex_ssw &= 0xceff; /* do not re-run instruction */
#else
	/* 
	 * The easiest way to prevent the 68040 from re-running the 
	 * instruction.
	 */
	efp->ex_pc = (unsigned long) rtn68040; 
#endif
}



/*----------------------------------------------------------------------
 * Function: probe()
 *
 * Probe address access for reading (flag < 0) or writing (flag >= 0).
 * User should use sizeof() to specify the access size.
 *
 * OUTPUT: 	. 0 if access is allowed
 * 		. 1 if not.
 *---------------------------------------------------------------------*/
extern int probe(addr, size, flag)
char *addr;
int size;
int flag;
{
	register int (*save_buserr)();
	int (**busvecp)();
	unsigned long ps;

	busvecp = (int(**)()) asm_catch_VBR();
	busvecp += OFF_BUSERR;

	debug(("probe 0x%p\n",addr));

	save_flags(ps);
	cli();

	save_buserr = *busvecp;
	caughtbuserr = 0;
	*busvecp = asm_catch_bus_error;

#ifdef MC68040
	cache_disable();	/* flush & disable cache */
	asm volatile ("nop");	/* flush the pipeline & pending write */
#endif

	if (flag < 0)
	{
		switch (size)
		{
		case sizeof(char):
			fakevalue = *addr;
			break;

		case sizeof(short):
			fakevalue = *(short *)addr;
			break;

		case sizeof(long):
			fakevalue = *(long *)addr;
			break;
		}
	}
	else
	{
		switch (size)
		{
		case sizeof(char):
			*addr = fakevalue;
			break;

		case sizeof(short):
			*(short *)addr = fakevalue;
			break;

		case sizeof(long):
			*(long *)addr = fakevalue;
			break;
		}
	}

#ifdef MC68040
	asm volatile ("rtn68040:");
#endif

	*busvecp = save_buserr;

#ifdef MC68040
	cache_enable();
#endif

	restore_flags(ps);

	return (caughtbuserr);
}



/*----------------------------------------------------------------------
 * Function: init_module()
 *
 * Initialize the module
 *---------------------------------------------------------------------*/

int init_module()
{
	return 0;
}



/*----------------------------------------------------------------------
 * Function: cleanup_module()
 *
 * Undid whatever init_module did
 *---------------------------------------------------------------------*/

void cleanup_module()
{
}
